/* We define some stuff in this file before the headers because
   the headers #define symbols used in these functions */

/* Verbose and somewhat useful malloc/free tracing */
/* turn it on by hand-munging the obvious spot in headers.h */

#ifdef TRACE_MALLOC
void *MYmalloc(size_t size) {
  void *result = malloc(size);
  fprintf(stderr, "Allocated %6d at %x\n", size, result);
  return result;
}
char *MYcalloc(size_t nmemb, size_t size) {
  char *result = calloc(nmemb,size);
  fprintf(stderr, "Allocated %6d at %x\n", nmemb*size, result);
  return result;
}
void MYfree(void *ptr){
  fprintf(stderr, "freeing %x\n", ptr);
  free(ptr);
}
#endif

/* A ridiculous global flag to insure we don't do a nc_redef on a file
   that's already in definition mode (why this is an error is a
   mystery to me */
int def_mode = 1;

/* and some functions to use for changing modes that maintain the state
   of the flag above, since there is apparently no netCDF API call to
   determine if you're in definition mode or not (!)
*/
int MYnc_redef(int ncid) {
  if(!def_mode) {
    def_mode = 1;
    return nc_redef(ncid);
  }
}

int MYnc_enddef(int ncid) {
  if(def_mode) {
    def_mode = 0;
    return nc_enddef(ncid);
  }
}

/* I guess we need to see the standard headers here, even though
   they're massive overkill (but then, massive header overkill is
   traditional in C programming, isn't it?
*/
#include "headers.h"

/* code support for SS error "handling" */
void spit_status(int stat, char *call) {
  if(stat < 0) {
    fprintf(stderr, "ERROR: code %d result of %s\n", stat, call);
    exit(-1);
  }
}


/* We use this to allocate string buffers because many of the HDF-EOS
   string-buffer-smashing routines do *NOT* null-terminate the strings
   they smash into the buffers.  I can see why not (FORTRAN compatibility),
   but it does make for counter-intuitive C processing...
*/
char *newstring(int len) {
  char *result = calloc(len, 1);
  if(result == NULL)
    spit_status(-1, "Ran out of memory allocating a little bitty string!");
  return result;
}

/* a little string copier - surely this is a library function somewhere? */
char *strclone(char *in)
{
  char *result = (char *) malloc(strlen(in) + 1);
  strcpy(result, in);
  return result;
}

/* next_name and end_name take pointers to a name buffer
   (one of those comma-seperated atrocities the HDF inquiry functions return).
   end_name checks to see if the current name is terminated by \0 or ',', and
   clobbers any ',' with \0.
   next_name just returns the pointer moved forward to the next field.
   This means you can write loops of the form:

   nameptr = namebuf;
   for(i=0; i< numnames; i++) {
     end_name(nameptr);
     ... do stuff with nameptr
     nameptr = next_name(nameptr);
   }

   avoiding the extra malloc/freeing needed with their breakout utility.
*/

void end_name(char *namebuf)
{
  while(*namebuf != 0 && *namebuf != ',') namebuf++;
  if(*namebuf == ',') *namebuf = 0;
}

char *next_name (char *namebuf)
{
  char *result = namebuf;
  if(namebuf == NULL || *namebuf == 0) return result;
  while(*result != 0 && *result != ',') result++;
  return ++result;
}


/* Dimension maps get passed around as a pointer to a pointer to
   the head of the map
*/

struct dment *find_dimension (struct dment *map, char *name)
{
  struct dment *result;
  for(result = map;
      result != NULL && strcmp(name, result->Hname);
      result = result->next);
  return result;
}

struct dment *add_dimension(struct dment **map, char *Hname, char *Nname, int size, int id)
{
  struct dment *foo = (struct dment *)malloc(sizeof (struct dment));
  foo->id = id;
  foo->size = size;
  foo->Hname = strclone(Hname);
  /* In the normal case, Hname and Nname will be the same, so let's
     avoid consing more storage in that case... */
  if(strcmp(Hname, Nname))
    foo->Nname = strclone(Nname);
  else
    foo->Nname = foo->Hname;
  foo->next = *map;
  *map = foo;
  return foo;
}

/* netCDF is pickier about variable names than HDF-EOS.  In particular,
   spaces are OK in the latter, but not the former.
*/
void fix_nc_varname(char *vn) {
  /* For now, we only 'fix' spaces.  There is a small chance that this won't
     work (if the file has a variable named "foo bar" and another named
     "foo_bar" there will be a collision).  If we need to generalize this,
     we'll do so later */
  char *spacept = vn;

  while((spacept = strpbrk(spacept, " ")) != NULL) {
    *spacept++ = '_';
  }
}
