/* support for converting swaths */

#include "headers.h"

/* a little enumerated type to control which field name space gets looked
   at in write_swath_fields */
typedef enum {
  DATA_SWATH = 0,
  GEO_SWATH = 1
} swath_space;

void write_swath_fields
(char *swname, hid_t swhandle, int ncid, struct dment **dmap, char *fieldnamebuf, int numfields, swath_space ss);

void move_swath_attributes(int swhandle, char *swname, int ncid);

void record_data_geolocation(hid_t swhandle, char *swname, int ncid, struct dment **dmap);

void convert_swath(char *hdfname, char *netcdfname, int ncid)
{
  /* To convert a swath file, we need to extract the dimensions of
     all its swaths and create them in the netCDF file, then pull out
     the data and attributes and recreate them on the netCDF side.
  */
  char *swnamebuf;
  char *swname;
  hid_t filehandle;
  int nswath;
  int swindex;
  struct dment *dmap = NULL;

  {
    /* get the number of swaths and their names */
    long strbufsize;
    SS(nswath = HE5_SWinqswath(hdfname, NULL, &strbufsize));
    swnamebuf = newstring(strbufsize + 1);
    SS(nswath = HE5_SWinqswath(hdfname, swnamebuf, &strbufsize));
  }

  SS(filehandle = HE5_SWopen(hdfname, H5F_ACC_RDONLY));

  TRACE((stderr, "File contains swath(s):\n\t|%s|\n", swnamebuf));
 
  /* walk the swaths, noting their dimensions and hosing
     their fields into the netCDF file.

     One big difference between HDF-EOS and netCDF is that in netCDF,
     dimensions are properties of a file, but in HDF-EOS they're
     properties of a swath (with multiple swaths/file possible).
     So, for each swath we walk its dimensions and look them up
     in a dimension map.  If there's an entry in the map and it doesn't
     match what we just looked up, we create a new outgoing netCDF dimension
     named {dimension}_{swathname}.  This breaks COADS compliance somewhat;
     sorry about that...
  */

  swname = swnamebuf;
  for(swindex=0; swindex<nswath; swindex++) {
    hid_t swhandle;
    int numdim = 0;
    long bufsize;
    hsize_t *dimsizebuf;
    char *dimnamebuf;
    char *dimname;
    int fieldindex;
    int numfields;
    char *fieldnamebuf = newstring(HE5_HDFE_NAMBUFSIZE);
    char *fieldname;
    int *dimidbuf;
    int totaldatasize;
    int i;

    end_name(swname);

    /* grab the swath handle */
    TRACE((stderr, "Attaching to swath |%s|\n", swname));
    SS(swhandle = HE5_SWattach(filehandle, swname));

    /* unroll its dimension names and sizes */
    SS(numdim = HE5_SWnentries(swhandle, HE5_HDFE_NENTDIM, &bufsize));

    dimnamebuf = newstring(bufsize + 1);
    dimsizebuf = (hsize_t *) malloc(sizeof(hsize_t) * numdim);
    SS(numdim = HE5_SWinqdims(swhandle, dimnamebuf, dimsizebuf));
    TRACE((stderr, "%d dimensions:\n\t|%s|\n", numdim, dimnamebuf));

    /* For each dimension, make an entry in the dimension map */
    dimname = dimnamebuf;
    for(i=0; i<numdim; i++) {
      struct dment *olddim = find_dimension(dmap, dimname);
      end_name(dimname);
      if(olddim == NULL) {
	/* First time we've seen this name - create an entry */
	int newdimid;
	TRACE((stderr, "defining dimension |%s|, size %d\n",
	       dimname, dimsizebuf[i]));
	/* Map HDF Unlimited to netCDF unlimited here */
	if(dimsizebuf[i] == H5S_UNLIMITED)
	  dimsizebuf[i] = NC_UNLIMITED;
	SS(nc_def_dim(ncid, dimname, dimsizebuf[i], &newdimid));
	add_dimension(&dmap, dimname, dimname, dimsizebuf[i], newdimid);
      }
      else {
	/* If we've seen this name before, but with a different size,
	   smash the old entry with a new netCDF name and size
	*/
	if(olddim->size != dimsizebuf[i]) {
	  char *newdimname;
	  TRACE((stderr, "Dimension %s has old size %d and new size %d)!\n",
		 dimname, olddim->size, dimsizebuf[i]));
	  newdimname = newstring(strlen(dimname) + strlen(swname) + 2);
	  sprintf(newdimname, "%s_%s", dimname, swname);
	  TRACE((stderr, "Creating new netCDF dimension %s\n", newdimname));
	  if(olddim->Nname != olddim->Hname)
	    free(olddim->Nname);
	  olddim->Nname = newdimname;
	  olddim->size = dimsizebuf[i];
	  SS(nc_def_dim(ncid, dimname, dimsizebuf[i], &(olddim->id)));
	}
      }
	
      dimname = next_name(dimname);
    }
    free(dimsizebuf);
    free(dimnamebuf);

    /* unroll its field names and data types */
    SS(numfields = HE5_SWnentries(swhandle, HE5_HDFE_NENTDFLD, &bufsize));
    if(numfields > 0) {
      fieldnamebuf = newstring(bufsize + 1);
      SS(numfields = HE5_SWinqdatafields(swhandle, fieldnamebuf, NULL, NULL));

      TRACE((stderr, "Attached to swath |%s| with %d data fields:\n\t|%s|\n",
	     swname, numfields, fieldnamebuf));
      
      write_swath_fields(swname, swhandle, ncid, &dmap, fieldnamebuf, numfields, DATA_SWATH);
      free(fieldnamebuf);
    }

    /* do the same for the geolocation fields */
    SS(numfields = HE5_SWnentries(swhandle, HE5_HDFE_NENTGFLD, &bufsize));
    if(numfields > 0) {
      fieldnamebuf = newstring(bufsize + 1);
      SS(numfields = HE5_SWinqgeofields(swhandle, fieldnamebuf, NULL, NULL));

      TRACE((stderr, "Attached to swath |%s| with %d geolocation fields:\n\t|%s|\n",
	     swname, numfields, fieldnamebuf));
      
      write_swath_fields(swname, swhandle, ncid, &dmap, fieldnamebuf, numfields, GEO_SWATH);
      free(fieldnamebuf);
    }
    /* move the attributes */
    move_swath_attributes(swhandle, swname, ncid);

    /* and record the data/geolocation mappings */
    record_data_geolocation(swhandle, swname, ncid, &dmap);

    HE5_SWdetach(swhandle);
    swname = next_name(swname);
  }
  free(swnamebuf);
  HE5_SWclose(filehandle);
  nc_close(ncid);
}


  /* OK, each field then maps to a variable name (swath_varname, to
     insure uniqueness, I guess) in the netCDF file */
char *swath_space_prefix[] = {"", "GeoLoc_"};
int swath_space_group[] = {HE5_HDFE_DATAGROUP, HE5_HDFE_GEOGROUP};

void write_swath_fields
(char *swname, hid_t swhandle, int ncid, struct dment **dmap, char *fieldnamebuf, int numfields, swath_space ss) {
  char *fieldname = fieldnamebuf;
  int fieldindex;
  int numdim;
  int *dimidbuf;
  hsize_t *dimsizebuf;
  char *dimnamebuf;
  char *dimname;
  int i;

  for(fieldindex=0; fieldindex< numfields; fieldindex++) {
    int fieldtype;
    long totalfieldentries = 1;
    void *databuf;
    int varid;
    char *varname;
    int nctype;

    end_name(fieldname);

    /* This is how to determine the real type of a field,
       the number of dimensions it has,
       and the real size of each dimension.  We can safely clobber
       the existing dimsizebuf and dimnamebuf because the dimensions for
       a field have to be a subset of the dimensions for its swath
       (if this isn't true, some HDF-EOS designer needs a swat with a
       clue-by-four).
    */
    SS(HE5_SWfieldinfo(swhandle, fieldname,
		       &numdim, NULL, &fieldtype, NULL, NULL));
    dimidbuf = (int *) malloc(numdim * sizeof(int));
    dimsizebuf = (hsize_t *)malloc(numdim * sizeof(hsize_t));

    { long dimnamebufsize = 0;
    SS(HE5_SWnentries(swhandle, HE5_HDFE_NENTDIM, &dimnamebufsize));
    TRACE((stderr, "Dimension name buf is %d bytes\n", dimnamebufsize));
    dimnamebuf = newstring(dimnamebufsize + 1);
    }
    /* dimnamebuf = newstring(numdim * 100); */
    SS(HE5_SWfieldinfo(swhandle, fieldname,
		       &numdim, dimsizebuf, &fieldtype, dimnamebuf, NULL));
    TRACE((stderr, "Processing field %s with %d dimensions:\n\t|%s|\n",
	   fieldname, numdim, dimnamebuf));

    /* compute the total number of elements in this swath */
    /* and build a dimension ID vector for variable creation later */
    dimname = dimnamebuf;
    for(i=0; i<numdim; i++) {
      struct dment *dimentry;
      end_name(dimname);
      dimentry = find_dimension(*dmap, dimname);
      if(dimentry == NULL) {
	/* A new (to us) dimension - map it.
	   There is apparently no reliable way to determine if
	   this dimension is unlimited - this API call returns
	   the actual size.  So, if this dimension was unlimited,
	   we won't record that in the netCDF file.  Pity...
	*/
	int newdimid;
	TRACE((stderr, "defining late dimension |%s|, size %d\n",
	       dimname, dimsizebuf[i]));
	SS(nc_def_dim(ncid, dimname, dimsizebuf[i], &newdimid));
	dimentry = add_dimension(dmap, dimname, dimname, dimsizebuf[i], newdimid);
      }
      else {
	/* If this dimension doesn't match what we looked up, we
	   can't translate this file without serious conniptions
	   (changing the name of the dimension, creating multiple
	   netCDF files), so we error out if that happens
	*/
	if(dimentry->size != dimsizebuf[i]) {
	  fprintf(stderr, "Dimension %s has two different sizes (%d and %d)!\n",
		  dimname, dimentry->size, dimsizebuf[i]);
	  spit_status(-1, "bad dimension sizes");
	}
      }
      TRACE((stderr, "Found dimension |%s|, id %d, size %d\n",
	     dimname, dimentry->id, dimentry->size));
      dimidbuf[i] = dimentry->id;
      totalfieldentries *= dimsizebuf[i];

      dimname = next_name(dimname);
    }
    TRACE((stderr, "totalfieldentries is %d\n", totalfieldentries));

    /* OK, determine the actual type and size and allocate a buffer
       for the field.
       Apparently the only way to determine the machine type is 
       by getting out the size (in bytes) of each element.  BLEAH!*/
    {
      hid_t typeID;
      H5T_class_t classID;
      H5T_order_t order;
      size_t tsize;

      /* NOTE: this is an undocumented HDF-EOS API function that is
	 apparently the only way to get the real type information out.
      */
      SS(HE5_SWinqdatatype(swhandle, fieldname, NULL, swath_space_group[ss],
			   &typeID, &classID, &order, &tsize));
      TRACE((stderr, "tsize is %d\n", tsize));

      databuf = (void *)malloc(totalfieldentries * tsize);
      if(databuf == NULL) {
	/* Yes, I could write arbitrarily hairy code to read/write the
	   field one slice at a time, but I'm feeling a little lazy today...
	*/
	fprintf(stderr, "Can't convert swath %s, field %s of type %d - can't allocate enough memory\n", swname, fieldname, fieldtype);
	/* NOTE again - skipping to the next field here */
	fieldname = next_name(fieldname);
	continue;
      }

      /* Here we map the HDF type to a netCDF type using the size
	 information we just extracted */
      switch(classID) {
      case H5T_INTEGER:
	switch(tsize) {
	case 1: nctype = NC_BYTE; break; /* NC_CHAR? */
	case 2: nctype = NC_SHORT; break;
	case 4: nctype = NC_INT; break;
	default: spit_status(-1, "Bogus size for integer number");
	}
	break;
      case H5T_FLOAT:
	switch(tsize) {
	case 4: nctype = NC_FLOAT; break;
	case 8: nctype = NC_DOUBLE; break;
	default: spit_status(-1, "Bogus size for floating-point number");
	}
	break;
      default:
	/* OK, let's explicitly throw this one away and just warn */
	fprintf(stderr, "Can't convert swath %s, field %s of type %d\n",
		swname, fieldname, fieldtype);
	/* NOTE: skipping to the next field here */
	fieldname = next_name(fieldname);
	continue;
      }
    }

    SS(HE5_SWreadfield(swhandle, fieldname, NULL, NULL, NULL, databuf));

    /* OK we've theoretically read the data in, now hose it back out */
    /* We construct variable names in the result file as
       swathname_fieldname, to insure they're globally unique.
       We may have to translate variable names here - netCDF variable
       names are pretty restrictive ( [a-zA-Z][0-9a-zA-Z_]* )
    */
    varname = newstring(strlen(fieldname)+strlen(swname) +strlen(swath_space_prefix[ss])+ 2);
    sprintf(varname, "%s%s_%s", swath_space_prefix[ss], swname, fieldname);
    fix_nc_varname(varname);

    /* netCDF files are moded, so for the convenience of this loop
       you get to switch into and out of define mode.  Bleah. */
    /* I hope this isn't a  major performance drag - if it is, I'll have
       to restructure this loop.  Bleah**2 */
    SS(nc_redef(ncid));

    SS(nc_def_var(ncid, varname, nctype, numdim, dimidbuf, &varid));
    /* if there are any attributes on this field, move them to this
       variable now while we're still in define mode */
    SS(nc_enddef(ncid));
    switch(nctype) {
    case NC_BYTE:
      SS(nc_put_var_uchar(ncid, varid, (unsigned char *)databuf));
      break;
    case NC_SHORT:
      SS(nc_put_var_short(ncid, varid, (short *)databuf));
      break;
    case NC_INT:
      SS(nc_put_var_int(ncid, varid, (int *)databuf));
      break;
    case NC_FLOAT:
      SS(nc_put_var_float(ncid, varid, (float *)databuf));
      break;
    case NC_DOUBLE:
      SS(nc_put_var_double(ncid, varid, (double *)databuf));
      break;
    default:
      fprintf(stderr, "Swath %s, field %s, netCDF type %d, ???\n",
	      swname, fieldname, nctype);
      spit_status(-1, "How the hell did we get here???");
      break;
    }
    /* There is still stuff to do here about moving the attributes */
    free(varname);
    free(databuf);
    fieldname = next_name(fieldname);
  }
}

/* this moves swath attributes to netCDF global attributes */
/* In addition to the inqattrs/readattr pair, there also exist
   inqgrpattrs/readgrpattr and inqlocattrs/readlocattr.  Problem is,
   there is only attrinfo, which you must have to get the real machine-local
   data type of attributes.  So, I'm going to punt on converting the
   grp and loc attributes.  Foo!
*/
void move_swath_attributes(int swhandle, char *swname, int ncid)
{
  char *attnamebuf;
  long attbufsize;
  char *attname;
  int attcount;
  int attindex;
  int nctype;
  /* just long enough for the simple scalars */
  char tempatt[9];

  /* we need to be in netCDF define mode here */
  nc_redef(ncid);

  SS(attcount = HE5_SWinqattrs(swhandle, NULL, &attbufsize));
  attnamebuf = newstring(attbufsize+1);
  SS(HE5_SWinqattrs(swhandle, attnamebuf, &attbufsize));
  
  TRACE((stderr, "Moving %d attributes of swath |%s|:\n\t|%s|\n",
	 attcount, swname, attnamebuf));
  attname = attnamebuf;
  for(attindex = 0; attindex < attcount; attindex++) {
    H5T_class_t atttype;
    int attsize;
    char *ncattname;
    void *attentbuf;
    hsize_t atteltcount;

    int tsize;

    end_name(attname);
    ncattname = newstring(strlen(swname)+strlen(attname)+2);
    sprintf(ncattname, "%s_%s", swname, attname);
    fix_nc_varname(ncattname);
    TRACE((stderr, "Attribute |%s| moves to |%s|\n", attname, ncattname));

    SS(HE5_SWattrinfo(swhandle, attname, &atttype, &atteltcount));
    TRACE((stderr, "Type number %d, count %d\n", atttype, atteltcount));
    
    { 
      hid_t typeID;
      H5T_class_t classID;
      H5T_order_t order;
      /* NOTE: this is an undocumented HDF-EOS API function that is
	 apparently the only way to get the real type information out.
      */
      SS(HE5_SWinqdatatype(swhandle, NULL, attname, HE5_HDFE_ATTRGROUP,
			   &typeID, &classID, &order, &tsize));
    }

    /* We have the number of attribute elements, their size, and their
       abstract type, which is what we need to map them to their
       netCDF concrete type.
    */
    switch(atttype) {
    case H5T_INTEGER:
      switch(tsize) {
      case 1: nctype = NC_BYTE; break;
      case 2: nctype = NC_SHORT; break;
      case 4: nctype = NC_INT; break;
      default: spit_status(-1, "Bogus size for integer number");
      }
      break;
    case H5T_FLOAT:
      switch(tsize) {
      case 4: nctype = NC_FLOAT; break;
      case 8: nctype = NC_DOUBLE; break;
      default: spit_status(-1, "Bogus size for floating-point number");
      }
      break;
    case H5T_STRING:
      /* Strings are a weird special case in HDF-EOS, as they are stored
	 not as vectors of char, but as scalars with a specific size, so
	 here, tsize IS the number of characters.  To make the rest of the
	 code work, we smack it to 1 here.
      */
      nctype = NC_CHAR;
      atteltcount = tsize;
      tsize = 1;
      break;
    default:
      fprintf(stderr, "Unknown type of attribute: %d\n", atttype);
      spit_status(-1, "Unhandled attribute type");
      break;
    }
    attsize = atteltcount * tsize;

    if(attsize < 9)
      attentbuf = (void *)tempatt;
    else
      attentbuf = newstring(attsize);
    SS(HE5_SWreadattr(swhandle, attname, attentbuf));

    switch(nctype) {
    case NC_BYTE:
      SS(nc_put_att_uchar(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (char *)attentbuf));
      break;
    case NC_CHAR:
      SS(nc_put_att_text(ncid, NC_GLOBAL, ncattname, atteltcount, (char *)attentbuf));
      break;
    case NC_SHORT:
      SS(nc_put_att_short(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (short *)attentbuf));
      break;
    case NC_INT:
      SS(nc_put_att_int(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (int *)attentbuf));
      break;
    case NC_FLOAT:
      SS(nc_put_att_float(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (float *)attentbuf));
      break;
    case NC_DOUBLE:
      SS(nc_put_att_double(ncid, NC_GLOBAL, ncattname, nctype, atteltcount, (double *)attentbuf));
      break;
    }
    /* Free the attribute memory */
    if(attentbuf != tempatt)
      free(attentbuf);
    attname = next_name(attname);
  }
}

/* Minor bit of nastiness here.  HDF-EOS gives us a dimension map, which is
    conceptually a per-swath list of (geodim, datadim, offset, increment)
    tuples.  netCDF has no way to store a vector of strings (dammit), so
    we store this information as three global attributes named:

    {swath}_DimensionMap
    {swath}_DimensionMapOffsets
    {swath}_DimensionMapIncrements

    We use the global dimension map to map the HDF-EOS dimension names
    to the netCDF names, when they're different.  This is rare, but it
    can happen...
 */
void record_data_geolocation(hid_t swhandle, char *swname, int ncid, struct dment **dmap) {
  char *atttail = "_DimensionMap";
  char *offtail = "_DimensionMapOffsets";
  char *inctail = "_DimensionMapIncrements";
  int mapcount;
  long mapsize;
  char *dimmap;
  long *offset;
  long *increment;
  int i;
  char *newmap;
  char *dimptr;
  char *newmapend;    
  char *dmattname;
  char *dmoffname;
  char *dmincname;


  SS(mapcount = HE5_SWnentries(swhandle, HE5_HDFE_NENTMAP, &mapsize));
  TRACE((stderr, "%d dimension maps, total size %d\n", mapcount, mapsize));
  if(mapcount > 0) {
    dimmap = newstring(mapsize + 1);
    offset = (long *) malloc(sizeof(long) * mapcount);
    increment = (long *) malloc(sizeof(long) * mapcount);

    SS(HE5_SWinqmaps(swhandle, dimmap, offset, increment));
    TRACE((stderr, "dimension map: |%s|\n", dimmap));
    for(i=0; i<mapcount; i++) {
      TRACE((stderr, "Offset: %d\tIncrement: %d\n", offset[i], increment[i]));
    }
    /* worst case for the remapped dimension map is current size +
       mapcount * (length of swathname + 1), so that's what we allocate
    */
    newmap = newstring(mapsize + mapcount * (strlen(swname) + 1));
    newmapend = newmap;

    dimptr = dimmap;
    for(i=0; i<mapcount; i++) {
      char *geodim;
      char *datadim;
      struct dment *geodment;
      struct dment *datadment;

      end_name(dimptr);

      /* slightly sleazy code here.  dimptr points to a string of
	 the form "geodim/datadim".  We want to look up "geodim" and
	 "datadim" seperately in the dimension map, so we find the "/"
	 in the string and smash it with a null.  Then we put the null
	 back so next_name will work properly.
      */
      geodim = dimptr;
      datadim = strchr(geodim, '/');
      *datadim = 0;
      datadim++;
      geodment = find_dimension(*dmap, geodim);
      datadment = find_dimension(*dmap, datadim);
      sprintf(newmapend, "%s/%s,", geodment->Nname, datadment->Nname);
      newmapend += strlen(newmapend);
      *(datadim-1) = '/';

      dimptr = next_name(dimptr);
    }
    /* The remapped dimension map has a spare comma at its end - clobber
       it here */
    newmap[strlen(newmap)-1] = 0;

    TRACE((stderr, "Remapped dimension map: |%s|\n", newmap));

    dmattname = newstring(strlen(swname) + strlen(atttail));
    dmoffname = newstring(strlen(swname) + strlen(offtail));
    dmincname = newstring(strlen(swname) + strlen(inctail));
    sprintf(dmattname, "%s%s", swname, atttail);
    sprintf(dmoffname, "%s%s", swname, offtail);
    sprintf(dmincname, "%s%s", swname, inctail);

    SS(nc_put_att_text(ncid, NC_GLOBAL, dmattname, strlen(newmap), newmap));
    SS(nc_put_att_long(ncid, NC_GLOBAL, dmoffname, NC_INT, mapcount, offset));
    SS(nc_put_att_long(ncid, NC_GLOBAL, dmincname, NC_INT, mapcount, increment));

    free(dmattname);
    free(dmincname);
    free(dmoffname);
    free(newmap);
    free(increment);
    free(offset);
    free(dimmap);
  }
}
